import org.junit.Test;

import java.math.BigDecimal;
import java.math.RoundingMode;

import static org.junit.Assert.assertEquals;

/**
 * Generic test class for implementing and testing a very basic rational number class
 *                _ _   _               _
 * __      ___ __(_) |_| |_ ___ _ __   | |__  _   _
 * \ \ /\ / / '__| | __| __/ _ \ '_ \  | '_ \| | | |
 *  \ V  V /| |  | | |_| ||  __/ | | | | |_) | |_| |
 *   \_/\_/ |_|  |_|\__|\__\___|_| |_| |_.__/ \__, |
 *                                            |___/
 *  ____                  __     __               _
 * / ___|_   _____ _ __   \ \   / /__   __ _  ___| |
 * \___ \ \ / / _ \ '_ \   \ \ / / _ \ / _` |/ _ \ |
 *  ___) \ V /  __/ | | |   \ V / (_) | (_| |  __/ |
 * |____/ \_/ \___|_| |_|    \_/ \___/ \__, |\___|_|
 *                                     |___/
 * Licensed under the GPLv2 License, Version 2.0 (the "License");
 * Copyright (c) Sven Vogel
 */
public class Aufgabe3 {

    /**
     * A 64-bit signed rational number. The nominator and denominator are of fixed size of 32-bit.
     * Both can be signed. The sign of the number is given by sign(nominator) * sign(denominator)
     * Zero can be represented as the nominator being zero and the denominator being anything but 0.
     * Since division by 0 is undefined, all rational numbers with a denominator of zero are to be treated as
     * invalid numbers.
     * @param nominator
     * @param denominator
     */
    private record Rational(int nominator, int denominator) {

        /**
         * Construct a rational number. Throws a {@link IllegalArgumentException} exception if the denominator is zero.
         * @param nominator the nominator
         * @param denominator the denominator (any non-zero value)
         * @apiNote nominator and denominator are GCD reduced to avoid overflows
         */
        private Rational(int nominator, int denominator) {
            if (denominator == 0)
                throw new IllegalArgumentException("denominator must be non-negative");

            // GCD reduction
            var gcd = gcd(denominator, nominator);

            this.nominator = nominator / gcd;
            this.denominator = denominator / gcd;
        }

        /**
         * compute the greatest common divisor of x and y
         *
         * @param x the first number to compute the greatest common divisor of
         * @param y the second number to compute the greatest common divisor of
         * @return the greatest common divisor of x and y
         */
        private static int gcd(int x, int y) {
            if (y == 0)
                return x;

            return gcd(y, y % x);
        }

        /**
         * Multiply this number by rhs. Yields the result as a new number
         * @param rhs the right hand side of the multiplication
         * @return the result of this * rhs
         * @apiNote nominator and denominator are GCD reduced to avoid overflows
         */
        public Rational multiply(Rational rhs) {
            return new Rational(this.nominator * rhs.nominator(), this.denominator * rhs.denominator());
        }

        /**
         * Add this number to rhs. Yields the result as a new number
         * @param rhs the right hand side of the sum
         * @return the result of this + rhs
         * @apiNote nominator and denominator are GCD reduced to avoid overflows
         */
        public Rational add(Rational rhs) {
            return new Rational(this.nominator * rhs.denominator() + this.denominator * rhs.nominator(), this.denominator * rhs.denominator());
        }

        /**
         * Returns an approximation of the represented number by converting the fraction to BigDecimals and performing
         * the conversion to single number by division.
         * @param scale the number of digits to approximate
         * @return the approximated value as BigDecimal
         */
        public BigDecimal toBigDecimal(int scale) {
            var nominator = BigDecimal.valueOf(this.nominator, scale);
            var denominator = BigDecimal.valueOf(this.denominator, scale);

            return nominator.divide(denominator, RoundingMode.HALF_EVEN);
        }
    }

    @Test
    public void test() {
        var a = new Rational(1, 3);
        var b = new Rational(7, 6);

        var c = a.add(b).multiply(a);

        assertEquals(c.toBigDecimal(10).compareTo(BigDecimal.valueOf(0.5)), 0);
    }
}